# -*- coding: utf-8 -*-
from  pymodbus.client.sync import ModbusTcpClient
import time

class ModbusClient:
    def __init__(self, devId, host, logfile):
        self.devId = devId
        self.host = host
        self.fieldFileName="campos.txt"
        self.idCampos = []
        self.tags = []
        self.messageError=[]
        self.log_path = logfile

#----------------------------------------------------------------------------
#ReadFieldFile
#Función de lectura que almacena los identificadores de los campos en un array
#Parametros de entrada: register y number(nº campos leidos=nReg/2)
#----------------------------------------------------------------------------
    def ReadFieldFile(self):
        file=open(self.fieldFileName,'r')
        self.idCampos= [ line.rstrip('\n') for line in file.readlines()]
        file.close()

#----------------------------------------------------------------------------
#parseRegistersGen
#Función de Parseo generica
#Parametros de entrada: register y number(nº campos leidos=nReg/2)
#----------------------------------------------------------------------------
    def parseRegistersGen(self, register, number):
        k = 0
        data = []
        for x in range(number):
            value=(register[k]<<16) + register[k+1]
            data.append(self.complement2(value,32))
            #data.append(((register[k]<<16) + register[k+1]))
            k += 2
        return data # devuelve una lista con los datos de los registros

#----------------------------------------------------------------------------
#complement2
#Función que realiza la conversion a valor con signo de un numero en complemento a 2
#Parametros de entrada: valor y nº de bits->necesario para comprobar si se ha realizado el complemento a 2
#----------------------------------------------------------------------------

    def complement2(self,data,nbits):
        if (data>>(nbits-1)) != 0 :
            signed_value = data if data < (1 << nbits-1) else data - (1 << nbits)
        else:
            signed_value = data
        return signed_value
#----------------------------------------------------------------------------
#ReadInputRegistersTcp
#Función para realizar la lectura de registros de entrada modbus función 04 del protocolo modbus
#Parametros de entrada: Dirección IP del servidor, Identificador de dispositivo, dirección de comienzo y nº de registros
#----------------------------------------------------------------------------
    def ReadInputRegistersTcp(self,server, devId, startAddress, numReg):#function 04 modbus
        client = ModbusTcpClient(server)
        data = client.read_input_registers(startAddress, numReg,unit=devId)
        client.close()
        if data.function_code > 0x80:#si se recive algun tipo de error en la lectura
            if self.tags.count('errorCode_'+str(data.function_code)) == 0:
                self.tags.append('errorCode_'+str(data.function_code))
            self.messageError.append('errorCode_'+str(data.function_code)+': Server='+server+' devId='+str(devId)+' startAddress='+str(startAddress)+' numReg='+str(numReg))


        return data.registers
#----------------------------------------------------------------------------
#ReadHoldingRegisterTcp
#Función para realizar la lectura de registros de entrada modbus función 04 del protocolo modbus
#Parametros de entrada: Dirección IP del servidor, Identificador de dispositivo, dirección de comienzo y nº de registros
#----------------------------------------------------------------------------
    def ReadHoldingRegistersTcp(self,server, devId, startAddress, numReg):#function 03 modbus

        client = ModbusTcpClient(server)
        data = client.read_holding_registers(startAddress, numReg,unit=devId)
        client.close()
        if data.function_code > 0x80:#si se recive algun tipo de error en la lectura
            if self.tags.count('errorCode_'+str(data.function_code)) == 0:
                self.tags.append('errorCode_'+str(data.function_code))
            self.messageError.append('errorCode_'+str(data.function_code)+': Server='+server+' devId='+str(devId)+' startAddress='+str(startAddress)+' numReg='+str(numReg))
        #data.registers --> Extrae una lista con el valor de los registros
        return data.registers



#----------------------------------------------------------------------------
#parseRegisterUnid
#Función de Parseo de datos especificos para primer bloque
#Parametros de entrada: register y number(nº campos leidos=nReg/2)
#----------------------------------------------------------------------------
    #Función especifica para rellenar la lista de instantaneos, maximo y minimos valores de circuitor de los primeros 30 valores
    def parseRegistersUnid(self, register, number):
        k = 0
        data = []
        """Bucle que rellena una lista con lso datos, de los registros, se unen pares registro continuo
        con los bits mas significativos en el primer registro de los dos y luego se multiplican por el valor
        correspondiente en función del valor
        """
        for i in range (number):
            value=(register[k]<<16) + register[k+1]
            if i == 0 :
                #data.append(float(((register[k]<<16) + register[k+1]))/10)
                data.append(float(self.complement2(value,32))/10)
            elif i==6 or i==7:
                data.append(float(self.complement2(value,32))/100)
                #data.append(float((register[k]<<16) + register[k+1]) / 100)
            else:
                data.append(float(self.complement2(value,32)))
                #data.append(float((register[k]<<16) + register[k+1]))
            k += 2
        return data

#----------------------------------------------------------------------------
#parseRegisterIII
#Función de Parseo de datos especificos para bloque de III
#Parametros de entrada: register y number(nº campos leidos=nReg/2)
#----------------------------------------------------------------------------
    def parseRegisterIII(self, register, number):
        k = 0
        data = []
        """Bucle que rellena una lista con los datos, de los registros(16bits), se unen pares registro continuo
        con los bits mas significativos en el primer registro de los dos(32bits) y luego se multiplican por el valor
        correspondiente en función del valor
        """
        for i in range (number):
            value=(register[k]<<16) + register[k+1]
            if i == 7 or i == 8 or i==9 or i==11 or i==12 or i==13 or i==14 or i==15 or i==16 :
                data.append(float(self.complement2(value,32))/10)
                #data.append(float(((register[k]<<16) + register[k+1]))/10)
            elif i==4 or i==5 or i==6 :
                data.append(float(self.complement2(value,32))/100)
                #data.append(float((register[k]<<16) + register[k+1]) / 100)
            else:
                data.append(float(self.complement2(value,32)))
                #data.append(float((register[k]<<16) + register[k+1]))
            k += 2
        return data

#----------------------------------------------------------------------------
#parseRegisterHarm
#Función de Parseo de datos especificos para Harmonicos
#Parametros de entrada: register, nreg y init,
#----------------------------------------------------------------------------
    def parseRegistersHarm(self, register, nreg, init):
        k = 0
        data = []
        for i in range (nreg):
            if i==0 and (init == 2600 or init == 2632 or init == 2664 or init == 2696 or init == 2728 or init == 2760):
                value=(register[k]<<16) + register[k+1]
                data.append(float(self.complement2(value,32))/10)
                #data.append(float((register[k]<<16) + register[k+1])/10)
            else:
                if init == 2602 or init == 2634 or init == 2666 or init == 2698 or init == 2730 or init == 2762:
                    data.append(float(self.complement2(register[i],16))/10)
                    #data.append(float(register[i])/10)
            k += 1
        return data

#----------------------------------------------------------------------------
#getInstantaneusRegister
#Función de lectura de registros instantaneos
#Parametros de entrada: -
#----------------------------------------------------------------------------
    def getInstantaneusRegister(self):
        listdata = []
        data = []
        readBlocks = [[0, 16], [16, 16],[32, 16], [48, 34], [2600, 2], [2602, 30], [2632, 2], [2634, 30], [2664, 2], [2666, 30], [2696, 2], [2698, 30], [2728, 2], [2730, 30], [2760, 2],[2762,30]]
        try:
            self.ReadFieldFile()
            # Dado que todos los valores no son continuos hay que realizar varias lecturas de diferentes secciones, para obtener los valores en el orden deseado

            for block in readBlocks:
                registers = self.ReadInputRegistersTcp(self.host, self.devId,block[0],block[1])#lectura de valores de equipo mediante modbusTCP
                if block[0] == 0 or block[0] == 16 or block[0] == 32:
                    data = data + self.parseRegistersUnid(registers,int(len(registers)/2))
                elif  block[0] == 48:
                    data = data + self.parseRegisterIII(registers,int(len(registers)/2))
                elif block[0] >=2600:
                    data = data + self.parseRegistersHarm(registers,int(len(registers)),block[0])
                else:
                    data = data + self.parseRegistersGen(registers,int(len(registers)/2))

            for i in range(len(data)):
                    listdata.append([self.idCampos[i], data[i]])
        except AttributeError:
            if self.tags.count('attributeError')==0:
                self.tags.append('attributeError')
            timestamp = time.strftime('%b  %d %H:%M:%s')
            filelog = open(self.log_path,'a')
            filelog.write(timestamp+':'+traceback.format_exc()+'\n')
            filelog.close()
        except Exception:
            if self.tags.count('unknowError')==0:
                self.tags.append('unknowError')
            timestamp = time.strftime('%b  %d %H:%M:%s')
            filelog = open(self.log_path,'a')
            filelog.write(timestamp+':'+traceback.format_exc()+'\n')
            filelog.close()
        listdata.append(['tags',self.tags])
        if len(self.messageError) > 0:
            listdata.append(['messageError',self.messageError])
        #self.tags=[]
        #self.messageError=[]

        return listdata
#----------------------------------------------------------------------------
#getSingleRegister
#Función de pruebas para la lectura de registros instantaneos
#Parametros de entrada: -
#----------------------------------------------------------------------------
    def getSingleRegister(self):
        listdata = []
        data = []
        readBlocks = [[4, 2],[60,2]]#,[6, 2],[8, 2],[12, 2],[20, 2],[22, 2],[24, 2],[28, 2],[36, 2],[38, 2],[40, 2],[44, 2],[48, 2],[50, 2],[52, 2],[56, 2]]
        self.ReadFieldFile()
        k=0
        try:
            for block in readBlocks:
                registers = self.ReadInputRegistersTcp(self.host, self.devId,block[0],block[1])#lectura de valores de equipo mediante modbusTCP
                """if block[0] == 0 or block[0] == 16 or block[0] == 32:
                    data = data + self.parseRegistersUnid(registers,int(len(registers)/2))
                elif  block[0] == 48:
                    data = data + self.parseRegisterIII(registers,int(len(registers)/2))
                elif block[0] >=2600:
                    data = data + self.parseRegistersHarm(registers,int(len(registers)),block[0])
                else:
                    data = data + self.parseRegistersGen(registers,int(len(registers)/2))"""
                print("Register1 ="+str(registers[0])+"\nRegister2="+str(registers[1]))
                prueba = hex((registers[0]<<16) + registers[1])
                print("Prueba hex = "+str(prueba))
                data = data + self.parseRegistersGen(registers,int(len(registers)/2))
                signed_value = self.complement2(data[k],32)
                print('Signed value= '+str(signed_value))
                k=k+1

            #-----------------------------------------------------------------
            for i in range(len(data)):
                    listdata.append([self.idCampos[i], data[i]])
        except AttributeError:
            if self.tags.count('attributeError')==0:
                self.tags.append('attributeError')
        listdata.append(['tags',self.tags])
        if len(self.messageError) > 0:
            listdata.append(['messageError',self.messageError])
        #self.tags=[]
        #self.messageError=[]


        return listdata


if __name__ == "__main__":
    import sys
    circutor = ModbusClient(10,'127.0.0.1')
    data = circutor.getInstantaneusRegister()
    print(data)
    print(type(data[0][1]))
    """data = circutor.getMaximumRegister()
    print(data)
    data = circutor.getMinimumRegister()
    print(data)"""
