//*********************************************************************
//*************  Copyright (C) 2008        OWASYS  ********************
//*********************************************************************
//**
//** The copyright to the computer programs here in is the property of
//** OWASYS The programs may be used and/or copied only with the
//** written permission from OWASYS or in accordance with the terms and
//** conditions stipulated in the agreement/contract under which the
//** programs have been supplied.
//**
//*********************************************************************
//******************** Module information *****************************
//**
//** Project:
//** Description:
//**
//** Filename:      LoadLibs.cpp
//** Owner:         Owasys
//** Creation date: 19/01/2006
//**
//*********************************************************************
//******************** Revision history *******************************
//** Revision date       Comments                           Responsible
//** -------- ---------- ---------------------------------- -----------
//** P1A      19/01/2006 Created                            owzabc
//*********************************************************************

#define __LOAD_LIBS_CPP

//-----------------------------------------------------------------//
//System Includes
//-----------------------------------------------------------------//
#include <stdio.h>
#include <dlfcn.h>
#include <stdlib.h>

//-----------------------------------------------------------------//
//User Includes
//-----------------------------------------------------------------//
#include "LoadLibs.h"

//-----------------------------------------------------------------//
//Defines
//-----------------------------------------------------------------//

//-----------------------------------------------------------------//
//Macros
//-----------------------------------------------------------------//

//-----------------------------------------------------------------//
//External variables
//-----------------------------------------------------------------//

//-----------------------------------------------------------------//
//External functions
//-----------------------------------------------------------------//

//-----------------------------------------------------------------//
//Private variables
//-----------------------------------------------------------------//


/* Functions definition */

//-----------------------------------------------------------------//
// Function: LoadExternalLibrary()
// Input Params:
//                *wLibraryName: library name string
//                **wLibHandle: pointer to the handler of the
//                              dynamic library
// Output Params:
// Description:
//    Loads a dynamic library from the file named wLibraryName.
//-----------------------------------------------------------------//
VOID LoadExternalLibrary( CHAR *wLibraryName, VOID **wLibHandle )
{
   *wLibHandle = dlopen( wLibraryName, RTLD_LAZY );
   if( !*wLibHandle ) {
      printf("No shared library %s found", wLibraryName);
      printf("%s\n", dlerror());
      exit(1);
   }
}

//-----------------------------------------------------------------//
// Function: UnloadExternalLibrary()
// Input Params:
//                *wLibHandle: pointer to the handler of the
//                             dynamic library
// Output Params:
// Description:
//    Unloads a dynamic library.
//-----------------------------------------------------------------//
INT UnloadExternalLibrary( VOID *wLibHandle )
{
   if( (dlclose( wLibHandle) ) != 0 ) {
      printf("UnloadExternalLibrary() error\n");
		exit(1);
	}
   printf("UnloadExternalLibrary() ok\n");
	return 0;
}

//-----------------------------------------------------------------//
// Function: LoadIOSFunctions()
// Input Params:
//                *wLibHandle: pointer to the handler of the
//                             dynamic library
// Output Params:
// Description:
//    Calls to dlsym to get the addresses where the functions of
//    the library are loaded to be used in the program.
//-----------------------------------------------------------------//
VOID LoadIOSFunctions( VOID *wLibHandle )
{
   FncIO_Initialize = ( INT ( *) ( VOID )) dlsym( wLibHandle, "IO_Initialize");
  	if( dlerror() != NULL ) {
      printf("No IO_Initialize() found\n");
   }
   FncIO_Start = ( INT ( *) ( VOID)) dlsym( wLibHandle, "IO_Start");
   if( dlerror() != NULL ) {
      printf("No IO_Start() found\n");
   }
   FncIO_Finalize = ( INT ( *) ( VOID)) dlsym( wLibHandle, "IO_Finalize");
   if( dlerror() != NULL ) {
      printf("No IO_Finalize() found\n");
   }
   FncIO_IsActive = ( INT ( *) ( INT *)) dlsym( wLibHandle, "IO_IsActive");
   if( dlerror() != NULL )  {
      printf("No IO_IsActive() found\n");
   }

   fncDIGIO_Enable_RS485 = ( INT ( *) ( VOID)) dlsym( wLibHandle, "DIGIO_Enable_RS485");
   if( dlerror() != NULL)  {
      printf("No DIGIO_Enable_RS485 found\n");
   }

   FncDIGIO_SetRx_RS485 = ( INT ( *) ( VOID )) dlsym( wLibHandle, "DIGIO_SetRx_RS485");
   if( dlerror() != NULL)  {
      printf("No DIGIO_SetRx_RS485 found");
   }
   FncDIGIO_SetTx_RS485 = ( INT ( *) ( VOID )) dlsym( wLibHandle, "DIGIO_SetTx_RS485");
   if( dlerror() != NULL)  {
      printf("No DIGIO_SetTx_RS485 found");
   }
}

//-----------------------------------------------------------------//
// Function: LoadRTUControlSelectFunctions()
// Input Params:
// Output Params:
// Description:
//-----------------------------------------------------------------//
VOID LoadRTUControlSelectFunctions( VOID *wLibHandle)
{
	FncRTUControl_Initialize = ( INT ( *) ( VOID *)) dlsym( wLibHandle, "RTUControl_Initialize");
  	if( dlerror() != NULL)  {
      printf("No RTUControl_Initialize() found\n");
   }
  	FncRTUControl_Start = ( INT ( *) ( VOID)) dlsym( wLibHandle, "RTUControl_Start");
  	if( dlerror() != NULL)  {
      printf("No RTUControl_Start() found\n");
   }
   FncRTUControl_Finalize = ( INT ( *) ( VOID)) dlsym( wLibHandle, "RTUControl_Finalize");
	if( dlerror() != NULL)  {
      printf("No RTUControl_Finalize() found\n");
  	}
}

//-----------------------------------------------------------------//
// Function: LoadRTUControlFunctions()
// Input Params:
//                *wLibHandle: pointer to the handler of the
//                             dynamic library
// Output Params:
// Description:
//    Calls to dlsym to get the addresses where the functions of
//    the library are loaded to be used in the program.
//-----------------------------------------------------------------//
VOID LoadRTUControlFunctions( VOID *wLibHandle)
{
   Fncusecsleep = ( INT ( *) (INT, INT)) dlsym( wLibHandle, "usecsleep");
   if( dlerror() != NULL)  {
      printf(" No usecsleep() found\n");
   }
}
