/*
 * File:   ModBus.cpp
 * Author: sarenet
 *
 * Created on 15 de junio de 2017, 8:20
 */
//libs necesarias carga e inicialización RTU, IO
#include <stdio.h>
#include <algorithm>
#include <cstdlib>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <fstream>
#include <ctime>
#include <vector>
#include <unistd.h>
#include "modbus/modbus-rtu.h"
#include "LoadLibs.h"
#include "owa3x/owerrors.h"
#include <jansson.h>
//Libreria MQTT
extern "C" {
#include "MQTTClient.h"
#include "MQTTClientPersistence.h"
}
//------------------------------------------------------------------------------
//Parametros de configuración Modbus
//------------------------------------------------------------------------------
#define RS485_PATH  "/dev/ttyS6"

//------------------------------------------------------------------------------
//Parametros de configurarción MQTT
//------------------------------------------------------------------------------
#define ADDRESS     "tcp://194.30.33.69:1883"
#define CLIENTID    "Owa3x"
#define QOS         1
#define TIMEOUT     10000L
//------------------------------------------------------------------------------
//Paths a ficheros y librerias
//------------------------------------------------------------------------------
#define FILE_FIELDS "/home/modbusApp/campos.txt"//path a fichero campos
#define FILE_LOG  "/var/log/Modbus_MQTT.log"//path a fichero log
#define LIBRTU    "/lib/libRTU_Module.so"
#define LIBIO     "/lib/libIOs_Module.so"

//#define NO_ERROR    0
#define TIME_REQUEST 60//Periodo en segundos entre una solicitudes de datos

#define LOCATION "Edificio103"


//-----------------------------------------------------------------------------
//Declaración de funciones
//----------------------------------------------------------------------------
using namespace std;
static int Enable_RS485();
int readInstantaneousData( modbus_t *);
int parseRegisterUnid(uint16_t *, int);
int parseRegisterIII(uint16_t *, int);
int parseRegisterGen(uint16_t *, int);
int parseRegisterHarm(uint16_t *, int,int);
int readFieldFile(char *);
int generateJSON(time_t, string);
void generateJSONData(time_t,string);
void checkError();
int writeLog(string, time_t);
int sendData(string,string);
//void pause(string);

//------------------------------------------------------------------------------
//Variables globales
//------------------------------------------------------------------------------
std::vector<float> datos;//vector donde se almacenan los datos leidos
std::vector<char*> campos;//Vector con los identificadores de los campos leidos en el fichero
string json;//JSON string, payload del MQTT
std::vector <char*> tags;
std::vector <int> error;
const string devId[] = {"Medidor_oficina-1","Medidor_General-1"};
int   slaveId[2] = {1,2};
json_t *payload = json_object();


//-------------------------------------------------------------------------------------------------------------
//Comunicación mediante Modbus, solicitud de datos y envio a traves de MQTT
//-------------------------------------------------------------------------------------------------------------

int main(int argc, char** argv) {
    modbus_t *ctx;
    time_t now;
    int rcIns,rcMax,rcMin;
    void *LibIOHandle = NULL;
    void *LibRTUControlHandle=NULL;
    int ReturnCode;
    //Load RTU e IOs Libraries
    LoadExternalLibrary((CHAR *)LIBRTU, &LibRTUControlHandle);
    LoadRTUControlFunctions(LibRTUControlHandle);
    LoadRTUControlSelectFunctions( LibRTUControlHandle);
    LoadExternalLibrary(LIBIO,&LibIOHandle);
    LoadIOSFunctions(LibIOHandle);

    //Initialize & Start IOs
    //pause("nInitialize RTU COntrol");
    printf("Initialize RTU\n");
    if( (ReturnCode = (*FncRTUControl_Initialize)(NULL)) != NO_ERROR) {
      now=time(0);
      writeLog("Error in RTUControl_Initialize()",now);
      UnloadExternalLibrary(LibRTUControlHandle);
      UnloadExternalLibrary(LibIOHandle);
      return 1;
   }
   //pause("Start RTU control");
   printf("Start RTU\n");
   if( (ReturnCode = (*FncRTUControl_Start)()) != NO_ERROR) {
      now=time(0);
      writeLog("Error in RTUControl_Start()",now);
      UnloadExternalLibrary(LibRTUControlHandle);
      UnloadExternalLibrary(LibIOHandle);
      return 1;
   }
   //pause("\nInitialize IOs");
   printf("Initialize IO\n");
    if(ReturnCode = (*FncIO_Initialize)() != NO_ERROR)
    {
        now=time(0);
        writeLog("Error in IO_Initialize",now);
        UnloadExternalLibrary(LibRTUControlHandle);
        UnloadExternalLibrary(LibIOHandle);
        return 1;
    }
   //pause("\nStart IOs");
   printf("Start IO\n");
    if( (ReturnCode = (*FncIO_Start)())!= NO_ERROR)
    {
        now=time(0);
        writeLog("Error in IO_Start",now);
        UnloadExternalLibrary(LibRTUControlHandle);
        UnloadExternalLibrary(LibIOHandle);
        return 1;
    }
    //pause("\nActivar Uart5");
    // Activar UART5 como ModBus
    printf("Enable RS485\n");
   if ( (ReturnCode = Enable_RS485()) != NO_ERROR){
      writeLog("Error in Enable_RS485",now);
      return 1;
   }

    //-------------------------------------------------------------------------
    //Inicialización conexion modbus
    //------------------------------------------------------------------------
    //pause("\n Establecer conexion Modbus");
    printf("New modbus RTU\n");
    ctx = modbus_new_rtu(RS485_PATH, 9600, 'N', 8, 1);
    if (ctx == NULL) {
        now=time(0);
        //fprintf(stderr, "Unable to create the libmodbus context\n");
        writeLog( "Error al crear conexion ModbusRTU",now);
        return -1;
    }
    int rs;
    //pause("\nSet Slave");
    int ndevice=sizeof(slaveId)/sizeof(slaveId[0]);
    while(true){
      for(int i=0;i<ndevice;i++)
      {
        modbus_flush(ctx);
        //Fijar identificador de esclavo con el que se va a comunicar
        printf("Set slave medidor oficina\n");
        rs=modbus_set_slave(ctx,slaveId[i]);
        if(rs==-1)
        {
    	     now=time(0);
            writeLog("Error al fijar slave",now);
            //fprintf(stderr, "Unable to create the libmodbus context\n");
        }
        //modbus_set_debug(ctx,TRUE); //Activa el modo debug de la libreria libmodbus
        struct timeval response_timeout;
        modbus_get_response_timeout(ctx, &response_timeout);
        response_timeout.tv_sec = 1;
        response_timeout.tv_usec = 500000;
        modbus_set_response_timeout(ctx, &response_timeout);
        modbus_set_byte_timeout(ctx, &response_timeout);
        //printf("Response time out (tv_sec), %d",response_timeout.tv_sec);
        //printf("\nResponse time out (tv_usec), %d",response_timeout.tv_usec);
        //pause("\nModbus connect");
        now=time(0);
        writeLog("Conectando a medidor de la oficina",now);
        if (modbus_connect(ctx) == -1) {
             now=time(0);
             writeLog("Fallo en la conexion Modbus",now);
            //fprintf(stderr, "Connection failed: %s\n", modbus_strerror(errno));
            return -1;
        }
        now=time(0);
        //definición de tipo de bus
        if( modbus_rtu_set_serial_mode(ctx,MODBUS_RTU_RS485)){
            now=time(0);
            writeLog("Set serial 485 error",now);
            //printf( "Serial 485 error: %d\n", errno);
        }

      //-------------------------------------------------------------------------
      //Lectura modbus
      //------------------------------------------------------------------------

          readFieldFile(FILE_FIELDS);
          now=time(0);
          readInstantaneousData(ctx);

          //-------------------------------------------------------------------------
          //Generación de cuerpo de mensaje y envio mqtt
          //------------------------------------------------------------------------
          generateJSONData(now, devId[i]);
          //printf("\n%s",json.c_str());
          //fflush(stdout);
          sendData(json,devId[i]);//envio de mensaje mqtt
          datos.clear();//Vaciar el vector de datos tras el envio de los datos
          campos.clear();//Vaciar vector de campos tras el envio
          tags.clear();
          error.clear();
          modbus_close(ctx);
        }


        sleep(TIME_REQUEST);
    }


    return 0;
}

/*void pause(string msg)
{
	char input[32];
	char *p;
	printf("%s\n",msg);
        p=fgets(input,32,stdin);
}*/

//-----------------------------------------------------------------------------
//
//Activar RS485 en Uart5
//
//-----------------------------------------------------------------------------
static INT Enable_RS485( VOID)
{
   INT ReturnCode = 0;
   time_t now;

   ReturnCode = (*fncDIGIO_Enable_RS485)();
   if( ReturnCode == NO_ERROR ) {
      now=time(0);
      writeLog( "RS485 enabled",now);
   } else {
      now=time(0);
      writeLog( "Error enabling RS485",now);
   }
   return ReturnCode;
}

//----------------------------------------------------------------------------
//readInstantaneousData
//Función de lectura de registros instantaneos
//Parametros de entrada: *ctx -> Puntero identificador modbus
//----------------------------------------------------------------------------
int readInstantaneousData(modbus_t *ctx)
{
    int addr[16]={0,16,32,48,2600,2602,2632,2634,2664,2666,2696,2698,2728,2730,2760,2762};
    int nreg[16]={16,16,16,34,2,30,2,30,2,30,2,30,2,30,2,30};
    time_t now;
    int rc;
    string log;
    char errorn[40];
    //string test="prueba";
    char buf[10];
    for(int i=0;i<16;i++)
    {
        now=time(0);
        uint16_t buffer[nreg[i]];
        //printf("\n Instantaneous Element addr:%d nreg:%d",addr[i],nreg[i]);
        rc=modbus_read_input_registers(ctx,addr[i],nreg[i],buffer);
        if (rc == -1) {
            error.push_back(errno);
            log="Error lectura registros entrada Instantaneos nreg=";
            sprintf(buf,"%d",i);
            log.append(buf);
            log.append(" ");
            sprintf(errorn,"%s",modbus_strerror(errno));
            log.append(" ErrorNo:");
            sprintf(buf,"%d",errno);
            log.append(buf);
            log.append(" Error saved on vector:");
            sprintf(buf,"%d",error.at((error.size()-1)));
            log.append(buf);
            log.append(errorn);
            writeLog(log,now);

        }
        else
        {
            if(i==0||i==1||i==2)
            {
                parseRegisterUnid(buffer,nreg[i]);//función de parseo especifica
            }
            else if(i==3)
            {
		            parseRegisterIII(buffer, nreg[i]);
            }
            else if(i>=4)
            {
                parseRegisterHarm(buffer,nreg[i],addr[i]);//función de parseo Harmonicos inicial
            }
            else
            {
                parseRegisterGen(buffer,nreg[i]);//funcion de parseo generica
            }
        }

    }

}


//----------------------------------------------------------------------------
//parseRegisterUnid
//Función de Parseo de datos especificos para primer bloque
//Parametros de entrada: *data y nreg, data-> datos leidos en read...Data, nreg--> numero de datos leidos
//----------------------------------------------------------------------------
int parseRegisterUnid(uint16_t * data, int nreg)
{
    float reg;
    int k=0;
    for(int i=0;i<(nreg/2);i++)
    {
        if(i==0)
        {
            reg=(float)((data[k]<<16)+data[k+1])/10;
            datos.push_back(reg);
        }
        else if(i==6|| i==7)
        {
            reg=(float)((data[k]<<16)+data[k+1])/100;
            datos.push_back(reg);
        }
        else
        {
            reg=(float)((data[k]<<16)+data[k+1]);
            datos.push_back(reg);
        }
        k=k+2;
    }
    return 0;
}

//----------------------------------------------------------------------------
//parseRegisterIII
//Función de Parseo de datos especificos para bloque de III
//Parametros de entrada: *data y nreg, data-> datos leidos en read...Data, nreg--> numero de datos leidos
//----------------------------------------------------------------------------
int parseRegisterIII(uint16_t * data, int nreg)
{
    float reg;
    int k=0;
    for(int i=0;i<(nreg/2);i++)
    {
        if(i==7 || i==8 || i==9 || i==11 || i==12|| i==13|| i==14|| i==15|| i==16)
        {
            reg=(float)((data[k]<<16)+data[k+1])/10;
            datos.push_back(reg);
        }
        else if(i==4|| i==5 || i==6)
        {
            reg=(float)((data[k]<<16)+data[k+1])/100;
            datos.push_back(reg);
        }
        else
        {
            reg=(float)((data[k]<<16)+data[k+1]);
            datos.push_back(reg);
        }
        k=k+2;
    }
    return 0;
}

//----------------------------------------------------------------------------
//parseRegisterGen
//Función de Parseo de datos generiaca, sin operar ningun parametro
//Parametros de entrada: *data y nreg, data-> datos leidos en read...Data, nreg--> numero de datos leidos
//----------------------------------------------------------------------------
int parseRegisterGen(uint16_t * data, int nreg)
{
    float reg;
    int k=0;
    for(int i=0;i<(nreg/2);i++)
    {

        reg=(float)((data[k]<<16)+data[k+1]);
        datos.push_back(reg);
        k=k+2;

    }
    return 0;
}

//----------------------------------------------------------------------------
//parseRegisterHarm
//Función de Parseo de datos especificos para Harmonicos
//Parametros de entrada: *data, nreg y addr, data-> datos leidos en read...Data, nreg--> numero de datos leidos
//----------------------------------------------------------------------------
int parseRegisterHarm(uint16_t * data, int nreg, int addr)
{
    float reg;
    int k=0;
    for(int i=0;i<nreg;i++)
    {
        if(addr==2600 || addr==2632 || addr==2664 || addr==2696 || addr==2728 || addr==2760)
        {
            reg=(float)((data[k]<<16)+data[k+1])/10;
            datos.push_back(reg);
            k=k+2;
            i=i+2;
        }
        else
        {
            reg=(float)(data[k])/10;
            datos.push_back(reg);
	           k=k+1;
        }

    }
    return 0;
}


//----------------------------------------------------------------------------
// readFieldFile
//Función que lee un fichero con los identificadores de los campos leidos
//y los almacena en un vector campos.
//Parametros de entrada: *fileName
//----------------------------------------------------------------------------
int readFieldFile(char * fileName) {
    string line;
    ifstream myfile(fileName);
    time_t now=time(0);
    if (myfile.is_open()) {
        while (getline(myfile, line)) {
            char *temp = new char[line.length() + 1];
            strcpy(temp, line.c_str());
            campos.push_back(temp);
        }
        myfile.close();
    }
    else
    {
        writeLog("Unable to open file",now);
    }
}

//----------------------------------------------------------------------------
//generateJSONData
//Función que construye un objeto JSON para enviarlo a traves de MQTT
//Parametros de entrada: timestamp, sello de tiempo para marcar el instante en el que se
//han recogido los datos
//----------------------------------------------------------------------------
void generateJSONData(time_t timestamp,string DEVID) {
    //UTC format in timestamp
    /*char* dt;
    tm *gmtm = gmtime(&timestamp);
    dt = asctime(gmtm);*/
    json_t *tag_arr = json_array();
    char buf[sizeof "2011-10-08T07:07:09Z"];
    strftime(buf, sizeof buf, "%FT%TZ", gmtime(&timestamp));
    if(error.size()>0){
      checkError();
    }
    //readFieldFile(FILE_FIELDS);
    char buffer[20];//[2000];
    char ID[3];
    printf("\ndatos: %d",datos.size());
    printf("\ncampos: %d",campos.size());
    //Inserción de campos de datos en JSON
    for(int i=0;i<datos.size();i++)
    {
        json_object_set_new( payload,campos.at(i),json_real(datos.at(i)));
    }
    //Inserción de sello de tiempo en JSON
    json_t *timeobj=json_object();
    json_object_set_new( timeobj,"time",json_string(buf));
    json_object_set_new( payload,"metadata",timeobj);
    //Inserción de tags en JSON
    if(tags.size()>0)
    {

        json_object_set_new( payload, "tags", tag_arr );
        for(int i=0;i<tags.size();i++)
        {

            json_array_append(tag_arr,json_string(tags.at(i)));
        }
    }
    else
    {
        json_object_set_new( payload, "tags", tag_arr );
    }
    //Inserción de identificador de dispositivo en JSON
    json_object_set_new( payload, "dev_id", json_string(DEVID.c_str()));

}

//----------------------------------------------------------------------------
//generateJSON
//Función que construye un objeto JSON para enviarlo a traves de MQTT
//Parametros de entrada: timestamp, sello de tiempo para marcar el instante en el que se
//han recogido los datos
//----------------------------------------------------------------------------
int generateJSON(time_t timestamp,string DEVID) {
    //UTC format in timestamp
    /*char* dt;
    tm *gmtm = gmtime(&timestamp);
    dt = asctime(gmtm);*/
    char buf[sizeof "2011-10-08T07:07:09Z"];
    strftime(buf, sizeof buf, "%FT%TZ", gmtime(&timestamp));
    if(error.size()>0){
      checkError();
    }
    json="{";
    //readFieldFile(FILE_FIELDS);
    char buffer[20];//[2000];
    char ID[3];
    printf("\ndatos: %d",datos.size());
    printf("\ncampos: %d",campos.size());
    for(int i=0;i<datos.size();i++)
    {
        if(i>0)
        {
            json.append(",");
        }
        json.append("\"");
        json.append(campos.at(i));
        json.append("\"");
        json.append(":");
        sprintf(buffer,"%f",datos.at(i));
        json.append(buffer);
    }
    //campo metadata.time
    json.append(",");
    json.append("\"");
    json.append("metadata");
    json.append("\"");
    json.append(":");
    json.append("{");
    json.append("\"");
    json.append("time");
    json.append("\"");
    json.append(":");
    json.append("\"");
    json.append(buf);
    json.append("\"");
    json.append("}");
    //Indicador de estado de lectura
    if(tags.size()>0)
    {
        json.append(",");
        json.append("\"");
        json.append("tags");
        json.append("\"");
        json.append(":");
        json.append("[");
        for(int i=0;i<tags.size();i++)
        {
            if(i>0)
            {
                json.append(",");
            }
            json.append("\"");
            json.append(tags.at(i));
            json.append("\"");
        }
      json.append("]");
    }
    else
    {
        json.append(",");
        json.append("\"");
        json.append("tags");
        json.append("\"");
        json.append(":");
        json.append("[");
        json.append("]");
    }
    //campo dev_id
    json.append(",");
    json.append("\"");
    json.append("dev_id");
    json.append("\"");
    json.append(":");
    json.append("\"");
    json.append(DEVID);
    json.append("\"");
    json.append("}");

}



//----------------------------------------------------------------------------
//checkError()
//Función que comprueba si se ha producido algun tipo de error en la lectura y genera un tag por cada tipo de error que se añadira al mensaje mqtt
//Parametros de entrada: -
//----------------------------------------------------------------------------
void checkError()
{
  bool etime= std::find(error.begin(), error.end(), 110)==error.end();
  bool ecrc= std::find(error.begin(), error.end(), 112345690)==error.end();
  bool edata= std::find(error.begin(), error.end(), EMBBADDATA)==error.end();
  if(std::find(error.begin(), error.end(), 110) != error.end()) {
    tags.push_back("timeout_error");
  }
  if(std::find(error.begin(), error.end(), 112345690) != error.end())
  {
    tags.push_back("crc_error");
  }
  if(std::find(error.begin(), error.end(), EMBBADDATA) != error.end())
  {
    tags.push_back("invalid_data_error");
  }
  if(etime==true && ecrc==true && edata==true)
  {
    tags.push_back("undefined_error");
  }

}

//----------------------------------------------------------------------------
//writeLog
//Función que escribe la fecha y un string en un fichero de Log
//Parametros de entrada:data y timestamp, data->string de infromación,
//timestamp->objeto time_t para marcar la fecha y hora del evento
//----------------------------------------------------------------------------
int writeLog(string data,time_t timestamp)
{
    char dt[sizeof "2011-10-08T07:07:09Z"];
    strftime(dt, sizeof dt, "%FT%TZ", gmtime(&timestamp));
    string line;
    line.append(dt);
    line.append(": ");
    line.append(data);
    std::ofstream ofs;
    ofs.open (FILE_LOG, std::ofstream::out | std::ofstream::app);
    ofs <<line;
    ofs << "\n";
    ofs.close();
    return 0;
}

//----------------------------------------------------------------------------
//sendData
//Función que manda un mensaje MQTT con los datos del json generado
//Parametros de entrada:msg, string con objeto json
//----------------------------------------------------------------------------
int sendData(string msg,string DEVID)
{
    /*printf("Esto es una prueba\n");
    char a[]="8.8.8.8";
    char b[]="Yo";
    char c[]="topic/prueba/1";
    char msg[]="{'prueba':5,'canal':'segundo'}";*/

    string log;
    MQTTClient client;
    MQTTClient_connectOptions conn_opts = MQTTClient_connectOptions_initializer;
    MQTTClient_message pubmsg = MQTTClient_message_initializer;
    MQTTClient_deliveryToken token;
    int rc;
    time_t now;

    MQTTClient_create(&client, ADDRESS, CLIENTID,MQTTCLIENT_PERSISTENCE_NONE, NULL);
    conn_opts.keepAliveInterval = 20;
    conn_opts.cleansession = 1;
    char buffer[10];
    if ((rc = MQTTClient_connect(client, &conn_opts)) != MQTTCLIENT_SUCCESS)
    {
        now=time(0);
        log="Failed to connect Broker MQTT, return code ";

        sprintf(buffer,"%d",rc);
        log.append(buffer);
        log.append("\n");
        writeLog(log,now);
        return -1;
    }
    char *info=(char *)json.c_str();
    pubmsg.payload = info;
    pubmsg.payloadlen = strlen(json.c_str());
    pubmsg.qos = QOS;
    pubmsg.retained = 0;
    string topic="";
    topic.append("Emeter");
    topic.append("/");
    topic.append(LOCATION);
    topic.append("/");
    topic.append(DEVID);
    MQTTClient_publishMessage(client, topic.c_str(), &pubmsg, &token);
    rc = MQTTClient_waitForCompletion(client, token, TIMEOUT);
    if(rc==0)//message published
    {
        now=time(0);
        log="Message delivered";
        writeLog(log,now);

    }
    else//error on publish
    {
        now=time(0);
        log="Error ocurred publishing mesage, data: ";
        log.append(msg);
        writeLog(log,now);
    }
    MQTTClient_disconnect(client, 10000);
    MQTTClient_destroy(&client);
    return rc;
}
