# -*- coding: utf-8 -*-
import paho.mqtt.client as mqtt
import json
import ssl
import time
import pprint

class Mqtt_client:
    def __init__(self):
        self.mqttBrokerHost = "194.30.33.69"
        self.mqttBrokerPort = 1883
        self.mqttBrokerUser = "<user name>"
        self.mqttBrokerPassword = "<password>"
        self.pathLog = "/var/log/mqtt_events.log"
        self.location = "Edificio103"
        self.clientBroker = mqtt.Client() #Cliente para publicar los datos a elasticSearch
        self.clientTTN = mqtt.Client() #cliente que recive los datos de TTN

    # La función a la que se llama cuando se recive un CONNACK como respuesta del servidor.
    def on_connectTTN(self, client, userdata, flags, rc):
        timestamp = time.strftime('%b  %d %H:%M:%s')
        filelog = open(self.pathLog,'a')
        filelog.write(timestamp+': Connected to TTN broker with result code '+str(rc)+'\n')
        filelog.close()
        # Se realiza la subscripción en on_connect() con lo que si se pierde la conexion
        # cuando se reconecta las subscripciones se renuevan.
        #client.subscribe("be7a0984/devices/+/+/+")
        client.subscribe([("be7a0984/devices/+/up",0),("be7a0984/devices/+/events/+",2)])

    def on_connect(self, client, userdata, flags, rc):
        timestamp = time.strftime('%b  %d %H:%M:%s')
        filelog = open(self.pathLog,'a')
        if rc == 0:
            filelog.write(timestamp+': Connected to MQTT broker (RC: '+str(rc)+')\n' )
        else:
            filelog.write(timestamp+': Connection to MQTT broker failed (RC: '+str(rc)+')\n')
        filelog.close()

    def on_disconnect(self, client, userdata, rc):
        timestamp = time.strftime('%b  %d %H:%M:%s')
        filelog = open(self.pathLog,'a')
        if rc != 0:
            filelog.write('Unexpected disconnect (RC: '+str(rc)+')\n')
        filelog.write(timestamp+':Disconnected from MQTT broker\n')
        filelog.close()
    # La función a la que se llama cuando se recive un mensaje PUBLISH del servidor.
    def on_message(self,client, userdata, msg):
        timestamp = time.strftime('%b  %d %H:%M:%s')
        topic = msg.topic.split('/')
        newTopic=''
        if topic[3] == 'up':
            mensaje = json.loads(str(msg.payload))
            #self.clientBroker.publish('RPi.Data', msg.payload, 1)
            if mensaje['port']==2:#Indica los datos de las balizas GPS
                mensaje['coordinates'] = [mensaje['payload_fields']['longitude'], mensaje['payload_fields']['latitude']]
                newTopic='GPS'+'/'+mensaje['dev_id']
            else:
                newTopic='AirQ'+'/'+self.location+'/'+mensaje['dev_id']
            self.clientBroker.publish(newTopic, json.dumps(mensaje), 1)
            filelog = open(self.pathLog,'a')
            filelog.write(timestamp+': Uplink Message \n')
            pprint.pprint(mensaje,filelog)
            filelog.close()
            #print(timestamp+': Uplink Message ' )
            #pprint.pprint(mensaje)
        elif topic[3] == 'events':
            #self.clientBroker.publish(msg.topic, msg.payload, 1)
            self.clientBroker.publish('RPi.Data', msg.payload, 1)
            evento = json.loads(str(msg.payload))
            filelog = open(self.pathLog,'a')
            filelog.write(timestamp+': Event from device'+evento["dev_eui"]+' to topic '+msg.topic+'\n')
            filelog.close()
            #print(timestamp+': Event from device '+evento["dev_eui"]+' to topic '+msg.topic)
        else:
            filelog = open(self.pathLog,'a')
            filelog.write(timestamp+':Topic not expected: '+msg.topic+'\n')
            filelog.close()
            #self.clientBroker.publish(msg.topic, msg.payload, 1)
            self.clientBroker.publish('RPi.Data', msg.payload, 1)

    def launch_client(self, host, username, password):
        self.clientBroker.on_connect = self.on_connect
        self.clientBroker.on_disconnect = self.on_disconnect
        self.clientBroker.connect(self.mqttBrokerHost, self.mqttBrokerPort, 60)#Puerto 1883(MQTT) sin utilizar TLS, si se utiliza TLS el puerto 8883

        self.clientTTN.username_pw_set(username, password)#TTN configuración de credenciales, username=Application Id y password = Application Key en la consola TTN
        self.clientTTN.on_connect = self.on_connectTTN
        self.clientTTN.on_disconnect = self.on_disconnect
        self.clientTTN.on_message = self.on_message

        #Configuración TLS para TTN
        self.clientTTN.tls_set('/home/root/mqtt_forwarder/mqtt-ca.pem', tls_version=ssl.PROTOCOL_TLSv1)

        #Conexion a cienteTTN
        self.clientTTN.connect(host, 8883, 60)#Puerto 1883(MQTT) sin TLS, con TLS puerto 8883

        # Comenzando los bucles del cliente que se conecta a TTN y el cliente que envia los datos haci elasticsearch
        self.clientBroker.loop_start()
        self.clientTTN.loop_start()

        while True:
            time.sleep(5)

if __name__ == "__main__":
    import sys
    instance = Mqtt_client()
    instance.launch_client(str(sys.argv[1]), str(sys.argv[2]),str(sys.argv[3]))
